# Exercises on continuous optimization

# Exercise 1

# a ----------

x<-c(1.77,-0.23,2.76,3.80,3.47,56.75,-1.34,4.24,-2.44,
     3.29,3.71,-2.40,4.53,-0.07,-1.05,-13.87,-2.53,
     -1.75,0.27,43.21)
n<- length(x)

boxplot(x)
dotchart(x)

# b ------------

# Log-likelihood
loglik <- function(theta,x){
	return(sum(log(dcauchy(x,location=theta))))
}

theta<- seq(-10,10,0.1)

N<-length(theta)
logL<-rep(0,N)
for(i in 1:N) logL[i]<- loglik(theta[i],x)	


# or, more compactly (and faster)
logL<-sapply(theta,loglik,x)

plot(theta,logL,type="l") # The log-likelihood has two modes

# c ---------

# First derivative of log-likelihood
dloglik <- function(theta,x) return(2*sum((x-theta)/((x-theta)^2+1)))

# Bisection method
bisection <-function(fun,dfun,a,b,epsi,...){
	theta<-(a+b)/2
	delta<-1
	while(delta>epsi){
		theta0<-theta
		if(dfun(a,x)*dfun(theta0,x)<=0) b<-theta0 else a<-theta0
		theta<-(a+b)/2
		delta<-abs(theta-theta0)/abs(theta0)
		print(c(a,b,delta))
	}
	return(list(objective=fun(theta,x),optimum=theta))
}

opt<-bisection(loglik,dloglik,-1,3,1e-6,x)

plot(theta,logL,type="l") 
points(opt$optimum,opt$objective)


# d --------

# Second derivative of log-likelihood
d2loglik <- function(theta,x){
	return(2*sum( ((x-theta)^2-1)/((x-theta)^2+1)^2))
}

# Newton's method
newton <-function(fun,dfun,d2fun,theta0,epsi,tmax,...){
	delta=1
	t<-0
	while((delta>epsi)&(t<=tmax)){
	  t<-t+1
		theta<-theta0-dfun(theta0,x)/d2fun(theta0,x)
		delta<-abs(theta-theta0)/abs(theta0)
		obj<-fun(theta,x)
		print(c(t,theta,obj,delta))
		theta0<-theta
	}
	return(list(objective=obj,optimum=theta,
	            derivative=dfun(theta,x),
	            derivative2=d2fun(theta,x)))
}
theta0<- 0
opt=newton(loglik,dloglik,d2loglik,theta0,1e-6,1000,x)

plot(theta,logL,type="l") 
points(opt$optimum,opt$objective,pch=22)

# e -------

opt <-optimize(f=loglik,lower=-2,upper=2,maximum=TRUE,x=x)

plot(theta,logL,type="l") 
points(opt$maximum,opt$objective,pch=23)


#------
# Exercise 2

# We write functions for the log likelihood, its first and second derivatives, and the Fisher information

loglik1<-function(theta,x){ # loglikelihood
  n<-length(x)
  return(sum(x)*log(theta)-n*log(-log(1-theta)) -sum(log(x)))
}

dloglik1<-function(theta,x){  # first derivative
  n<-length(x)
  return(sum(x)/theta + n/((1-theta)* log(1-theta)))
}

d2loglik1<-function(theta,x){ # second derivative
  n<-length(x)
  return(-sum(x)/theta^2 + n*(1+log(1-theta))/((1-theta)^2* log(1-theta)^2))
}

fisher.info <-function(theta,x){ # Fisher information with minus sign
  n<-length(x)
  EX<- -1/log(1-theta) * theta/(1-theta)
  return(-n*EX/theta^2 + n*(1+log(1-theta))/((1-theta)^2* log(1-theta)^2))
}

x<-c(1, 1, 1, 1, 1, 1, 2, 2, 2, 3)
theta<- seq(0,1,0.01)
logL<-sapply(theta,loglik1,x)
dlogL<-sapply(theta,dloglik1,x)

par(mfrow=c(1,2))
plot(theta,logL,type="l") 
plot(theta,dlogL,type="l") 
abline(0,1)
par(mfrow=c(1,1))

# Newton
theta0<- 0.8
opt=newton(loglik1,dloglik1,d2loglik1,theta0,1e-6,1000,x)
par(mfrow=c(1,2))
plot(theta,logL,type="l") 
points(opt$optimum,opt$objective)
plot(theta,dlogL,type="l") 
points(opt$optimum,0)
abline(0,1)
par(mfrow=c(1,1))

# Fisher

d2logL<-sapply(theta,d2loglik1,x)
fish<-sapply(theta,fisher.info,x)
plot(theta,d2logL,type='l')
lines(theta,fish,lty=2)

theta0<- 0.8
opt=newton(loglik1,dloglik1,fisher.info,theta0,1e-6,1000,x)
par(mfrow=c(1,2))
plot(theta,logL,type="l") 
points(opt$optimum,opt$objective)
plot(theta,dlogL,type="l") 
points(opt$optimum,0)
abline(h=0)
par(mfrow=c(1,1))